// ******************************************************************
//
//
//						CVideo.cpp
//
//							By
//
//						Mike Dailly
//
//					 Copyright (c)1999
//					All rights reserved
//
//
// This is the DirectDraw module. It controls all aspects of DirectDraw 
// from screen modes to enumeration, from flipping to clearing.
//
// While this file is attached to the Minus4 project, it may be used in
// other external projects. However, credit must be given to the author(s).
//
//
// Version	Date		Comments
// ----------------------------
//	1.0		13/3/99		First fully functional version. Open, Close, 
//						Clear, Flip etc..
//
//
// ******************************************************************
#include	<cvideo.h>
#include	<math.h>					// needed for POW() in SetGamma()

static	int32		NumScreenModes;		// Number of screen modes found
static	int32		DeviceNumber;		// Number of devices attached

// ******************************************************************
// Get string Length
// ******************************************************************
int	Len( char *pStr )
{
	int i = 0;
	while (*pStr++!=0x0) i++;
	return i+1;
}

// ******************************************************************
//
// return a string based on the last error
//
// ******************************************************************
char *CVideo::GetError( int hResult )
{
    LPTSTR lpszError;

    switch (hResult)
    {
    case DD_OK:
        // The request completed successfully.
        lpszError = TEXT("DD_OK");
        break;

    case DDERR_ALREADYINITIALIZED:
        // The object has already been initialized.
        lpszError = TEXT("DDERR_ALREADYINITIALIZED");
        break;

    case DDERR_BLTFASTCANTCLIP:
        // A DirectDrawClipper object is attached to a source surface 
        // that has passed into a call to the IDirectDrawSurface2::BltFast method. 
        lpszError = TEXT("DDERR_BLTFASTCANTCLIP");
        break;

    case DDERR_CANNOTATTACHSURFACE:
        // A surface cannot be attached to another requested surface. 
        lpszError = TEXT("DDERR_CANNOTATTACHSURFACE");
        break;

    case DDERR_CANNOTDETACHSURFACE:
        // A surface cannot be detached from another requested surface. 
        lpszError = TEXT("DDERR_CANNOTDETACHSURFACE");
        break;

    case DDERR_CANTCREATEDC:
        // Windows cannot create any more device contexts (DCs). 
        lpszError = TEXT("DDERR_CANTCREATEDC");
        break;

    case DDERR_CANTDUPLICATE:
        // Primary and 3D surfaces, or surfaces that are 
        // implicitly created, cannot be duplicated. 
        lpszError = TEXT("DDERR_CANTDUPLICATE");
        break;

    case DDERR_CANTLOCKSURFACE:
        // Access to this surface is refused because an 
        // attempt was made to lock the primary surface without DCI support. 
        lpszError = TEXT("DDERR_CANTLOCKSURFACE"); 
        break;

    case DDERR_CANTPAGELOCK:
        // An attempt to page lock a surface failed. 
        // Page lock will not work on a display-memory 
        // surface or an emulated primary surface.
        lpszError = TEXT("DDERR_CANTPAGELOCK"); 
        break;

    case DDERR_CANTPAGEUNLOCK:
        // An attempt to page unlock a surface failed. 
        // Page unlock will not work on a display-memory 
        // surface or an emulated primary surface. 
        lpszError = TEXT("DDERR_CANTPAGEUNLOCK");
        break;

    case DDERR_CLIPPERISUSINGHWND:
        // An attempt was made to set a clip list for a DirectDrawClipper 
        // object that is already monitoring a window handle. 
        lpszError = TEXT("DDERR_CLIPPERISUSINGHWND");
        break;

    case DDERR_COLORKEYNOTSET:
        // No source color key is specified for this operation
        lpszError = TEXT("DDERR_COLORKEYNOTSET");
        break;

    case DDERR_CURRENTLYNOTAVAIL:
        // No support is currently available. 
        lpszError = TEXT("DDERR_CURRENTLYNOTAVAIL");
        break;

    case DDERR_DCALREADYCREATED:
        // A device context (DC) has already been returned for this surface. 
        // Only one DC can be retrieved for each surface. 
        lpszError = TEXT("DDERR_DCALREADYCREATED");
        break;

    case DDERR_DIRECTDRAWALREADYCREATED:
        // A DirectDraw object representing this driver 
        // has already been created for this process. 
        lpszError = TEXT("DDERR_DIRECTDRAWALREADYCREATED");
        break;

    case DDERR_EXCEPTION:
        // An exception was encountered while 
        // performing the requested operation. 
        lpszError = TEXT("DDERR_EXCEPTION");
        break;

    case DDERR_EXCLUSIVEMODEALREADYSET:
        // An attempt was made to set the cooperative 
        // level when it was already set to exclusive. 
        lpszError = TEXT("DDERR_EXCLUSIVEMODEALREADYSET");
        break;

    case DDERR_GENERIC:
        // There is an undefined error condition. 
        lpszError = TEXT("DDERR_GENERIC");
        break;

    case DDERR_HEIGHTALIGN:
        // The height of the provided rectangle 
        // is not a multiple of the required alignment. 
        lpszError = TEXT("DDERR_HEIGHTALIGN");
        break;

    case DDERR_HWNDALREADYSET:
        // The DirectDraw cooperative level window 
        // handle has already been set. It cannot 
        // be reset while the process has surfaces or palettes created. 
        lpszError = TEXT("DDERR_HWNDALREADYSET");
        break;

    case DDERR_HWNDSUBCLASSED:
        // DirectDraw is prevented from restoring state because the
        // DirectDraw cooperative level window handle has been subclassed. 
        lpszError = TEXT("DDERR_HWNDSUBCLASSED");
        break;

    case DDERR_IMPLICITLYCREATED:
        // The surface cannot be restored because 
        // it is an implicitly created surface. 
        lpszError = TEXT("DDERR_IMPLICITLYCREATED");
        break;
 
    case DDERR_INCOMPATIBLEPRIMARY:
        // The primary surface creation request 
        // does not match with the existing primary surface. 
        lpszError = TEXT("DDERR_INCOMPATIBLEPRIMARY");
        break;

    case DDERR_INVALIDCAPS:
        // One or more of the capability bits 
        // passed to the callback function are incorrect. 
        lpszError = TEXT("DDERR_INVALIDCAPS");
        break;

    case DDERR_INVALIDCLIPLIST:
        // DirectDraw does not support the provided clip list.  
        lpszError = TEXT("DDERR_INVALIDCLIPLIST");
        break;

    case DDERR_INVALIDDIRECTDRAWGUID:
        // The globally unique identifier (GUID) passed to the
        // DirectDrawCreate function is not a valid DirectDraw driver identifier. 
        lpszError = TEXT("DDERR_INVALIDDIRECTDRAWGUID");
        break;

    case DDERR_INVALIDMODE:
        // DirectDraw does not support the requested mode. 
        lpszError = TEXT("DDERR_INVALIDMODE");
        break;

    case DDERR_INVALIDOBJECT:
        // DirectDraw received a pointer that was an invalid DirectDraw object. 
        lpszError = TEXT("DDERR_INVALIDOBJECT");
        break;

    case DDERR_INVALIDPARAMS:
        // One or more of the parameters passed to the method are incorrect. 
        lpszError = TEXT("DDERR_INVALIDPARAMS");
        break;

    case DDERR_INVALIDPIXELFORMAT:
        // The pixel format was invalid as specified. 
        lpszError = TEXT("DDERR_INVALIDPIXELFORMAT");
        break;

    case DDERR_INVALIDPOSITION:
        // The position of the overlay on the destination is no longer legal. 
        lpszError = TEXT("DDERR_INVALIDPOSITION");
        break;

    case DDERR_INVALIDRECT:
        // The provided rectangle was invalid. 
        lpszError = TEXT("DDERR_INVALIDRECT");
        break;

    case DDERR_INVALIDSURFACETYPE:
        // The requested operation could not be performed
        // because the surface was of the wrong type. 
        lpszError = TEXT("DDERR_INVALIDSURFACETYPE");
        break;

    case DDERR_LOCKEDSURFACES:
        // One or more surfaces are locked, 
        // causing the failure of the requested operation. 
        lpszError = TEXT("DDERR_LOCKEDSURFACES");
        break;

    case DDERR_MOREDATA:
        // There is more data available than the specified 
        // buffer size could hold.
        lpszError = TEXT("DDERR_MOREDATA");
        break;

    case DDERR_NO3D:
        // No 3D hardware or emulation is present. 
        lpszError = TEXT("DDERR_NO3D");
        break;

    case DDERR_NOALPHAHW:
        // No alpha acceleration hardware is present or available, 
        // causing the failure of the requested operation. 
        lpszError = TEXT("DDERR_NOALPHAHW");
        break;

    case DDERR_NOBLTHW:
        // No blitter hardware is present. 
        lpszError = TEXT("DDERR_NOBLTHW");
        break;

    case DDERR_NOCLIPLIST:
        // No clip list is available. 
        lpszError = TEXT("DDERR_NOCLIPLIST");
        break;

    case DDERR_NOCLIPPERATTACHED:
        // No DirectDrawClipper object is attached to the surface object. 
        lpszError = TEXT("DDERR_NOCLIPPERATTACHED");
        break;

    case DDERR_NOCOLORCONVHW:
        // The operation cannot be carried out because 
        // no color-conversion hardware is present or available. 
        lpszError = TEXT("DDERR_NOCOLORCONVHW");
        break;

    case DDERR_NOCOLORKEY:
        // The surface does not currently have a color key. 
        lpszError = TEXT("DDERR_NOCOLORKEY");
        break;

    case DDERR_NOCOLORKEYHW:
        // The operation cannot be carried out because there 
        // is no hardware support for the destination color key. 
        lpszError = TEXT("DDERR_NOCOLORKEYHW");
        break;

    case DDERR_NOCOOPERATIVELEVELSET:
        // A create function is called without the 
        // IDirectDraw2::SetCooperativeLevel method being called. 
        lpszError = TEXT("DDERR_NOCOOPERATIVELEVELSET");
        break;

    case DDERR_NODC:
        // No DC has ever been created for this surface. 
        lpszError = TEXT("DDERR_NODC");
        break;

    case DDERR_NODDROPSHW:
        // No DirectDraw raster operation (ROP) hardware is available. 
        lpszError = TEXT("DDERR_NODDROPSHW");
        break;

    case DDERR_NODIRECTDRAWHW:
        // Hardware-only DirectDraw object creation is not possible; 
        // the driver does not support any hardware. 
        lpszError = TEXT("DDERR_NODIRECTDRAWHW");
        break;

    case DDERR_NODIRECTDRAWSUPPORT:
        // DirectDraw support is not possible with the current display driver. 
        lpszError = TEXT("DDERR_NODIRECTDRAWSUPPORT");
        break;

    case DDERR_NOEMULATION: 
        // Software emulation is not available. 
        lpszError = TEXT("DDERR_NOEMULATION");
        break;

    case DDERR_NOEXCLUSIVEMODE:
        // The operation requires the application to have 
        // exclusive mode, but the application does not have exclusive mode. 
        lpszError = TEXT("DDERR_NOEXCLUSIVEMODE");
        break;

    case DDERR_NOFLIPHW: 
        // Flipping visible surfaces is not supported. 
        lpszError = TEXT("DDERR_NOFLIPHW");
        break;

    case DDERR_NOGDI: 
        // No GDI is present. 
        lpszError = TEXT("DDERR_NOGDI");
        break;

    case DDERR_NOHWND: 
        // Clipper notification requires a window handle, 
        // or no window handle has been previously set 
        // as the cooperative level window handle. 
        lpszError = TEXT("DDERR_NOHWND");
        break;

    case DDERR_NOMIPMAPHW: 
        // The operation cannot be carried out because no 
        // mipmap texture mapping hardware is present or available. 
        lpszError = TEXT("DDERR_NOMIPMAPHW");
        break;

    case DDERR_NOMIRRORHW: 
        // The operation cannot be carried out because 
        // no mirroring hardware is present or available. 
        lpszError = TEXT("DDERR_NOMIRRORHW");
        break;

    case DDERR_NONONLOCALVIDMEM: 
        // An attempt was made to allocate non-local video memory
        // from a device that does not support non-local video memory.
        lpszError = TEXT("DDERR_NONONLOCALVIDMEM");
        break;
 
    case DDERR_NOOVERLAYDEST: 
        // The IDirectDrawSurface2::GetOverlayPosition method 
        // is called on an overlay that the IDirectDrawSurface2::UpdateOverlay 
        // method has not been called on to establish a destination. 
        lpszError = TEXT("DDERR_NOOVERLAYDEST");
        break;

    case DDERR_NOOVERLAYHW: 
        // The operation cannot be carried out because 
        // no overlay hardware is present or available. 
        lpszError = TEXT("DDERR_NOOVERLAYHW");
        break;

    case DDERR_NOPALETTEATTACHED: 
        // No palette object is attached to this surface. 
        lpszError = TEXT("DDERR_NOPALETTEATTACHED");
        break;

    case DDERR_NOPALETTEHW: 
        // There is no hardware support for 16- or 256-color palettes. 
        lpszError = TEXT("DDERR_NOPALETTEHW");
        break;

    case DDERR_NORASTEROPHW: 
        // The operation cannot be carried out because 
        // no appropriate raster operation hardware is present or available. 
        lpszError = TEXT("DDERR_NORASTEROPHW");
        break;

    case DDERR_NOROTATIONHW: 
        // The operation cannot be carried out because 
        // no rotation hardware is present or available. 
        lpszError = TEXT("DDERR_NOROTATIONHW");
        break;

    case DDERR_NOSTRETCHHW: 
        // The operation cannot be carried out because 
        // there is no hardware support for stretching. 
        lpszError = TEXT("DDERR_NOSTRETCHHW");
        break;

    case DDERR_NOT4BITCOLOR: 
        // The DirectDrawSurface object is not using a 
        // 4-bit color palette and the requested operation 
        // requires a 4-bit color palette. 
        lpszError = TEXT("DDERR_NOT4BITCOLOR");
        break;

    case DDERR_NOT4BITCOLORINDEX: 
        // The DirectDrawSurface object is not using a 4-bit 
        // color index palette and the requested operation 
        // requires a 4-bit color index palette. 
        lpszError = TEXT("DDERR_NOT4BITCOLORINDEX");
        break;

    case DDERR_NOT8BITCOLOR: 
        // The DirectDrawSurface object is not using an 8-bit 
        // color palette and the requested operation requires 
        // an 8-bit color palette. 
        lpszError = TEXT("DDERR_NOT8BITCOLOR");
        break;

    case DDERR_NOTAOVERLAYSURFACE: 
        // An overlay component is called for a non-overlay surface. 
        lpszError = TEXT("DDERR_NOTAOVERLAYSURFACE");
        break;

    case DDERR_NOTEXTUREHW: 
        // The operation cannot be carried out because no 
        // texture-mapping hardware is present or available. 
        lpszError = TEXT("DDERR_NOTEXTUREHW");
        break;

    case DDERR_NOTFLIPPABLE: 
        // An attempt has been made to flip a surface that cannot be flipped. 
        lpszError = TEXT("DDERR_NOTFLIPPABLE");
        break;

    case DDERR_NOTFOUND: 
        // The requested item was not found. 
        lpszError = TEXT("DDERR_NOTFOUND");
        break;

    case DDERR_NOTINITIALIZED: 
        // An attempt was made to call an interface method of a DirectDraw object 
        // created by CoCreateInstance before the object was initialized. 
        lpszError = TEXT("DDERR_NOTINITIALIZED");
        break;

    case DDERR_NOTLOCKED: 
        // An attempt is made to unlock a surface that was not locked. 
        lpszError = TEXT("DDERR_NOTLOCKED");
        break;

    case DDERR_NOTPAGELOCKED: 
        // An attempt is made to page unlock a surface 
        // with no outstanding page locks. 
        lpszError = TEXT("DDERR_NOTPAGELOCKED");
        break;

    case DDERR_NOTPALETTIZED: 
        // The surface being used is not a palette-based surface. 
        lpszError = TEXT("DDERR_NOTPALETTIZED");
        break;

    case DDERR_NOVSYNCHW: 
        // The operation cannot be carried out because 
        // there is no hardware support for vertical blank synchronized operations. 
        lpszError = TEXT("DDERR_NOVSYNCHW");
        break;

    case DDERR_NOZBUFFERHW: 
        // The operation to create a z-buffer in display memory 
        // or to perform a blit using a z-buffer cannot be carried 
        // out because there is no hardware support for z-buffers. 
        lpszError = TEXT("DDERR_NOZBUFFERHW");
        break;

    case DDERR_NOZOVERLAYHW: 
        // The overlay surfaces cannot be z-layered based 
        // on the z-order because the hardware does not 
        // support z-ordering of overlays. 
        lpszError = TEXT("DDERR_NOZOVERLAYHW");
        break;

    case DDERR_OUTOFCAPS: 
        // The hardware needed for the requested operation has already been allocated. 
        lpszError = TEXT("DDERR_OUTOFCAPS");
        break;

    case DDERR_OUTOFMEMORY: 
        // DirectDraw does not have enough memory to perform the operation. 
        lpszError = TEXT("DDERR_OUTOFMEMORY");
        break;

    case DDERR_OUTOFVIDEOMEMORY: 
        // DirectDraw does not have enough display memory to perform the operation. 
        lpszError = TEXT("DDERR_OUTOFVIDEOMEMORY");
        break;

    case DDERR_OVERLAYCANTCLIP: 
        // The hardware does not support clipped overlays. 
        lpszError = TEXT("DDERR_OVERLAYCANTCLIP");
        break;

    case DDERR_OVERLAYCOLORKEYONLYONEACTIVE: 
        // An attempt was made to have more than one color key active on an overlay. 
        lpszError = TEXT("DDERR_OVERLAYCOLORKEYONLYONEACTIVE");
        break;

    case DDERR_OVERLAYNOTVISIBLE: 
        // The IDirectDrawSurface2::GetOverlayPosition method is called on a hidden overlay. 
        lpszError = TEXT("DDERR_OVERLAYNOTVISIBLE");
        break;

    case DDERR_PALETTEBUSY: 
        // Access to this palette is refused 
        // because the palette is locked by another thread. 
        lpszError = TEXT("DDERR_PALETTEBUSY");
        break;

    case DDERR_PRIMARYSURFACEALREADYEXISTS: 
        // This process has already created a primary surface. 
        lpszError = TEXT("DDERR_PRIMARYSURFACEALREADYEXISTS");
        break;

    case DDERR_REGIONTOOSMALL: 
        // The region passed to the 
        // IDirectDrawClipper::GetClipList method is too small. 
        lpszError = TEXT("DDERR_REGIONTOOSMALL");
        break;

    case DDERR_SURFACEALREADYATTACHED: 
        // An attempt was made to attach a surface to 
        // another surface to which it is already attached. 
        lpszError = TEXT("DDERR_SURFACEALREADYATTACHED");
        break;

    case DDERR_SURFACEALREADYDEPENDENT: 
        // An attempt was made to make a surface a dependency 
        // of another surface to which it is already dependent. 
        lpszError = TEXT("DDERR_SURFACEALREADYDEPENDENT");
        break;

    case DDERR_SURFACEBUSY: 
        // Access to the surface is refused because the 
        // surface is locked by another thread. 
        lpszError = TEXT("DDERR_SURFACEBUSY");
        break;

    case DDERR_SURFACEISOBSCURED: 
        // Access to the surface is refused 
        // because the surface is obscured. 
        lpszError = TEXT("DDERR_SURFACEISOBSCURED");
        break;

    case DDERR_SURFACELOST: 
        // Access to the surface is refused because the 
        // surface memory is gone. The DirectDrawSurface 
        // object representing this surface should have 
        // the IDirectDrawSurface2::Restore method called on it. 
        lpszError = TEXT("DDERR_SURFACELOST");
        break;

    case DDERR_SURFACENOTATTACHED: 
        // The requested surface is not attached. 
        lpszError = TEXT("DDERR_SURFACENOTATTACHED");
        break;

    case DDERR_TOOBIGHEIGHT: 
        // The height requested by DirectDraw is too large. 
        lpszError = TEXT("DDERR_TOOBIGHEIGHT");
        break;

    case DDERR_TOOBIGSIZE: 
        // The size requested by DirectDraw is too large. 
        // However, the individual height and width are OK.
        lpszError = TEXT("DDERR_TOOBIGSIZE");
        break;

    case DDERR_TOOBIGWIDTH: 
        // The width requested by DirectDraw is too large. 
        lpszError = TEXT("DDERR_TOOBIGWIDTH");
        break;

    case DDERR_UNSUPPORTED: 
        // The operation is not supported. 
        lpszError = TEXT("DDERR_UNSUPPORTED");
        break;

    case DDERR_UNSUPPORTEDFORMAT: 
        // The FourCC format requested is not supported by DirectDraw. 
        lpszError = TEXT("DDERR_UNSUPPORTEDFORMAT");
        break;

    case DDERR_UNSUPPORTEDMASK: 
        // The bitmask in the pixel format requested is not supported by DirectDraw. 
        lpszError = TEXT("DDERR_UNSUPPORTEDMASK");
        break;

    case DDERR_UNSUPPORTEDMODE: 
        // The display is currently in an unsupported mode. 
        lpszError = TEXT("DDERR_UNSUPPORTEDMODE");
        break;

    case DDERR_VERTICALBLANKINPROGRESS: 
        // A vertical blank is in progress. 
        lpszError = TEXT("DDERR_VERTICALBLANKINPROGRESS");
        break;

    case DDERR_WASSTILLDRAWING: 
        // The previous blit operation that is transferring 
        // information to or from this surface is incomplete. 
        lpszError = TEXT("DDERR_WASSTILLDRAWING");
        break;

    case DDERR_WRONGMODE:
        // This surface cannot be restored because it was created in a different mode. 
        lpszError = TEXT("DDERR_WRONGMODE");
        break;

    case DDERR_XALIGN:
        // The provided rectangle was not horizontally aligned on a required boundary. 
        lpszError = TEXT("DDERR_XALIGN");
        break;

    case DDERR_VIDEONOTACTIVE:
        // The video port is not active
        lpszError = TEXT("DDERR_VIDEONOTACTIVE");
        break;


        //
        // D3D Immediate Mode Errors
        //

    case D3DERR_BADMAJORVERSION:
        // ???
        lpszError = TEXT("D3DERR_BADMAJORVERSION");
        break;

    case D3DERR_BADMINORVERSION:
        // ???
        lpszError = TEXT("D3DERR_BADMINORVERSION");
        break;

    case D3DERR_INVALID_DEVICE:
        // ???
        lpszError = TEXT("D3DERR_INVALID_DEVICE");
        break;

    case D3DERR_INITFAILED:
        // ???
        lpszError = TEXT("D3DERR_INITFAILED");
        break;

    case D3DERR_EXECUTE_CREATE_FAILED:
        // ???
        lpszError = TEXT("D3DERR_EXECUTE_CREATE_FAILED");
        break;

    case D3DERR_EXECUTE_DESTROY_FAILED:
        // ???
        lpszError = TEXT("D3DERR_EXECUTE_DESTROY_FAILED");
        break;

    case D3DERR_EXECUTE_LOCK_FAILED:
        // ???
        lpszError = TEXT("D3DERR_EXECUTE_LOCK_FAILED");
        break;

    case D3DERR_EXECUTE_UNLOCK_FAILED:
        // ???
        lpszError = TEXT("D3DERR_EXECUTE_UNLOCK_FAILED");
        break;

    case D3DERR_EXECUTE_LOCKED:
        // ???
        lpszError = TEXT("D3DERR_EXECUTE_LOCKED");
        break;

    case D3DERR_EXECUTE_NOT_LOCKED:
        // ???
        lpszError = TEXT("D3DERR_EXECUTE_NOT_LOCKED");
        break;

    case D3DERR_EXECUTE_FAILED:
        // ???
        lpszError = TEXT("D3DERR_EXECUTE_FAILED");
        break;

    case D3DERR_EXECUTE_CLIPPED_FAILED:
        // ???
        lpszError = TEXT("D3DERR_CLIPPED_FAILED");
        break;

    case D3DERR_TEXTURE_NO_SUPPORT:
        // ???
        lpszError = TEXT("D3DERR_TEXTURE_NO_SUPPORT");
        break;

    case D3DERR_TEXTURE_CREATE_FAILED:
        // ???
        lpszError = TEXT("D3DERR_TEXTURE_CREATE_FAILED");
        break;

    case D3DERR_TEXTURE_DESTROY_FAILED:
        // ???
        lpszError = TEXT("D3DERR_TEXTURE_DESTROY_FAILED");
        break;

    case D3DERR_TEXTURE_LOCK_FAILED:
        // ???
        lpszError = TEXT("D3DERR_TEXTURE_LOCK_FAILED");
        break;

    case D3DERR_TEXTURE_UNLOCK_FAILED:
        // ???
        lpszError = TEXT("D3DERR_TEXTURE_UNLOCK_FAILED");
        break;

    case D3DERR_TEXTURE_LOAD_FAILED:
        // ???
        lpszError = TEXT("D3DERR_LOAD_FAILED");
        break;

    case D3DERR_TEXTURE_SWAP_FAILED:
        // ???
        lpszError = TEXT("D3DERR_SWAP_FAILED");
        break;

    case D3DERR_TEXTURE_LOCKED:
        // ???
        lpszError = TEXT("D3DERR_TEXTURE_LOCKED");
        break;

    case D3DERR_TEXTURE_NOT_LOCKED:
        // ???
        lpszError = TEXT("D3DERR_TEXTURE_NOT_LOCKED");
        break;

    case D3DERR_TEXTURE_GETSURF_FAILED:
        // ???
        lpszError = TEXT("D3DERR_TEXTURE_GETSURF_FAILED");
        break;

    case D3DERR_MATRIX_CREATE_FAILED:
        // ???
        lpszError = TEXT("D3DERR_MATRIX_CREATE_FAILED");
        break;

    case D3DERR_MATRIX_DESTROY_FAILED:
        // ???
        lpszError = TEXT("D3DERR_MATRIX_DESTROY_FAILED");
        break;

    case D3DERR_MATRIX_SETDATA_FAILED:
        // ???
        lpszError = TEXT("D3DERR_MATRIX_SETDATA_FAILED");
        break;

    case D3DERR_MATRIX_GETDATA_FAILED:
        // ???
        lpszError = TEXT("D3DERR_MAXTRIX_GETDATA_FAILED");
        break;

    case D3DERR_SETVIEWPORTDATA_FAILED:
        // ???
        lpszError = TEXT("D3DERR_SETVIEWPORTDATA_FAILED");
        break;

    case D3DERR_INVALIDCURRENTVIEWPORT:
        // ???
        lpszError = TEXT("D3DERR_INVALIDCURRENTVIEWPORT");
        break;
	
    case D3DERR_INVALIDPRIMITIVETYPE:
        // ???
        lpszError = TEXT("D3DERR_INVALIDPRIMITIVETYPE");
        break;

    case D3DERR_INVALIDVERTEXTYPE:
        // ???
        lpszError = TEXT("D3DERR_INVALIDVERTEXTYPE");
        break;

    case D3DERR_TEXTURE_BADSIZE:
        // ???
        lpszError = TEXT("D3DERR_TEXTURE_BADSIZE");
        break;

    case D3DERR_MATERIAL_CREATE_FAILED:
        // ???
        lpszError = TEXT("D3DERR_MATERIAL_CREATE_FAILED");
        break;

    case D3DERR_MATERIAL_DESTROY_FAILED:
        // ???
        lpszError = TEXT("D3DERR_MATERIAL_DESTROY_FAILED");
        break;

    case D3DERR_MATERIAL_SETDATA_FAILED:
        // ???
        lpszError = TEXT("D3DERR_MATERIAL_SETDATA_FAILED");
        break;

    case D3DERR_MATERIAL_GETDATA_FAILED:
        // ???
        lpszError = TEXT("D3DERR_MATERIAL_GETDATA_FAILED");
        break;

    case D3DERR_INVALIDPALETTE:
        // ???
        lpszError = TEXT("D3DERR_INVALIDPALETTE");
        break;

    case D3DERR_ZBUFF_NEEDS_SYSTEMMEMORY:
        // ???
        lpszError = TEXT("D3DERR_ZBUFF_NEEDS_SYSTEMMEMORY");
        break;

    case D3DERR_ZBUFF_NEEDS_VIDEOMEMORY:
        // ???
        lpszError = TEXT("D3DERR_ZBUFF_NEEDS_VIDEOMEMORY");
        break;

    case D3DERR_SURFACENOTINVIDMEM:
        // ???
        lpszError = TEXT("D3DERR_SURFACENOTINVIDMEM");
        break;

    case D3DERR_LIGHT_SET_FAILED:
        // ???
        lpszError = TEXT("D3DERR_LIGHT_SET_FAILED");
        break;

    case D3DERR_LIGHTHASVIEWPORT:
        // ???
        lpszError = TEXT("D3DERR_LIGHTHASVIEWPORT");
        break;

    case D3DERR_LIGHTNOTINTHISVIEWPORT:
        // ???
        lpszError = TEXT("D3DERR_LIGHTNOTINTHISVIEWPORT");
        break;

    case D3DERR_SCENE_IN_SCENE:
        // ???
        lpszError = TEXT("D3DERR_SCENE_IN_SCENE");
        break;

    case D3DERR_SCENE_NOT_IN_SCENE:
        // ???
        lpszError = TEXT("D3DERR_SCENE_NOT_IN_SCENE");
        break;

    case D3DERR_SCENE_BEGIN_FAILED:
        // ???
        lpszError = TEXT("D3DERR_SCENE_BEGIN_FAILED");
        break;

    case D3DERR_SCENE_END_FAILED:
        // ???
        lpszError = TEXT("D3DERR_SCENE_END_FAILED");
        break;

    case D3DERR_INBEGIN:
        // ???
        lpszError = TEXT("D3DERR_INBEGIN");
        break;

    case D3DERR_NOTINBEGIN:
        // ???
        lpszError = TEXT("D3DERR_NOTINBEGIN");
        break;

    case D3DERR_NOVIEWPORTS:
        // ???
        lpszError = TEXT("D3DERR_NOVIEWPORTS");
        break;

    case D3DERR_VIEWPORTDATANOTSET:
        // ???
        lpszError = TEXT("D3DERR_VIEWPORTDATANOTSET");
        break;

    case D3DERR_VIEWPORTHASNODEVICE:
        // ???
        lpszError = TEXT("D3DERR_VIEWPORTHASNODEVICE");
        break;

    

    default:
        // Unknown DD/D3D/App Error
        lpszError = TEXT("UNKNOWN ERROR");
        break;
    }

    return lpszError;
}


#ifndef __NT_DDRAW__
//******************************************************************************************************************
//******************************************************************************************************************
//******************************************************************************************************************
//******************************************************************************************************************


// ******************************************************************
//
//	Create a Device class
//
// ******************************************************************
CDevice::CDevice( char *pName, char *pDesc )
{
	memset( &m_GUID, 0, sizeof(GUID) );			// clear GUID (this means primary display)

	m_pGUID	= &m_GUID;
	m_pNext = NULL;
	m_pDriverName = (char*) MALLOC( Len(pName) + Len(pDesc) );
	m_pDriverDescription = m_pDriverName + Len(pName);
	m_pFirstMode = NULL;
	m_pLastMode	 = NULL;

	memcpy( m_pDriverName, pName, Len(pName) );
	memcpy( m_pDriverDescription, pDesc, Len(pDesc) );
}


// ******************************************************************
//
//	delete device class, and free
//
// ******************************************************************
CDevice::~CDevice( void )
{
	FREE( m_pDriverName );
}


// ******************************************************************
// Release the direct draw handles IF we have one.
// ******************************************************************
void CVideo::ReleaseDDrawHandle( void )
{
	// If we have an IDirectDraw4 Handle, free it!
	if (m_DDraw.pDD4 != NULL)
	{
		m_DDraw.pDD4->Release();			// DDraw 2 handle
		m_DDraw.pDD4 = NULL;
	}


	// If we have an IDirectDraw Handle, free it!
	if (m_DDraw.pDD != NULL)
	{
		m_DDraw.pDD->Release();			// free handle
		m_DDraw.pDD = NULL;
	}
	m_CurrentDevice = 0;				// clear current device
}


// ******************************************************************
// Get direct draw handle of a passed device
// ******************************************************************
int CVideo::GetDDrawHandle( CDevice *pDevice )
{
	ReleaseDDrawHandle();			// release handle if we HAVE one!
	

	m_err = DirectDrawCreate(pDevice->m_pGUID, &m_DDraw.pDD, 0);
	if (m_err != DD_OK ) 
	{
		m_pErr = GetError( m_err );
		return MD_ERROR;
	}

	// Upgrade to IDirectDraw4 (DX6) interface
	m_err = m_DDraw.pDD->QueryInterface( IID_IDirectDraw4, (void**) &m_DDraw.pDD4);
	if (m_err != DD_OK ) 
	{
		m_pErr = GetError( m_err );
		return MD_ERROR;
	}

	m_CurrentDevice = pDevice->m_DeviceNumber;
	return MD_OKAY;
}

// ******************************************************************
// Get the BITS in a mask
// ******************************************************************
uint32 GetNumberOfBits( DWORD dwMask )
{
    uint32 wBits = 0;
    while( dwMask )
    {
        dwMask = dwMask & ( dwMask - 1 );  
        wBits++;
    }
    return wBits;
}



// ******************************************************************
//
//	Create a screen mode class. Also takes a surface and extracts
//	Shift info, x and y sizes and BPP.
//
// ******************************************************************
CScreenMode::CScreenMode( LPDDSURFACEDESC2 lpDDSurfaceDesc )
{
	uint32	bits,downcount,bitcount,bitmask;

	m_xsize = lpDDSurfaceDesc->dwWidth;
	m_ysize = lpDDSurfaceDesc->dwHeight;
	m_bpp	= lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount;	// get bits in screen - RGB bits per pixel (4, 8, 16, 24, or 32). 
	m_pNext = NULL;

	m_AlphaAND   = 0;
	m_AlphaShift = 0;


	// Is this an RGB mode?
	if (((lpDDSurfaceDesc->ddpfPixelFormat.dwFlags&DDPF_RGB)==0) OR 
		(lpDDSurfaceDesc->ddpfPixelFormat.dwRBitMask == 0))
	{
		m_RedAND		= 0;
		m_RedShift		= 0;
		m_GreenAND		= 0;
		m_GreenShift	= 0;
		m_BlueAND		= 0;
		m_BlueShift		= 0;
		return;
	}

	// RED
	bits = lpDDSurfaceDesc->ddpfPixelFormat.dwRBitMask; 
	downcount = 0;
	while ((bits&1)==0)
	{
		bits = bits>>1;
		downcount++;
	}
	bitmask = bits;
	bitcount = GetNumberOfBits(bits);

	m_RedAND		= bitmask<<(16+(8-bitcount));
	m_RedShift		= (16+(8-bitcount)) - downcount;


	// GREEN
	bits = lpDDSurfaceDesc->ddpfPixelFormat.dwGBitMask; 
	downcount = 0;
	while ((bits&1)==0)
	{
		bits = bits>>1;
		downcount++;
	}
	bitmask = bits;
	bitcount = GetNumberOfBits(bits);

	m_GreenAND		= bitmask<<(8+(8-bitcount));
	m_GreenShift	= (8+(8-bitcount)) - downcount;


	// BLUE
	bits = lpDDSurfaceDesc->ddpfPixelFormat.dwBBitMask; 
	downcount = 0;
	while ((bits&1)==0)
	{
		bits = bits>>1;
		downcount++;
	}
	bitmask = bits;
	bitcount = GetNumberOfBits(bits);

	m_BlueAND		= bitmask<<(8-bitcount);
	m_BlueShift		= (8-bitcount) - downcount;


	m_AlphaAND	 = 0;
	m_AlphaShift = 0;
}


// ******************************************************************
//
//	Screen Mode(s) Enumeration.
//
// ******************************************************************
HRESULT WINAPI EnumModesCallback( LPDDSURFACEDESC2 lpDDSurfaceDesc,    LPVOID lpContext )
{
	CDevice		*pDevice = (CDevice *) lpContext;
	CScreenMode	*pScreen;


	pScreen = new CScreenMode( lpDDSurfaceDesc );
	pScreen->m_Device = pDevice->m_DeviceNumber;		// set device number
	pScreen->m_ModeNumber = NumScreenModes;				// give made up screen number
	NumScreenModes++;

	if (pDevice->m_pFirstMode==NULL)
	{
		pDevice->m_pFirstMode = pDevice->m_pLastMode = pScreen;
	}
	else 
	{
		pDevice->m_pLastMode->m_pNext = pScreen;
		pDevice->m_pLastMode = pScreen;
	}

	return DDENUMRET_OK;		// carry on...
}


// ******************************************************************
//
//	Display Driver enumeration.
//
// ******************************************************************
BOOL FAR PASCAL EnumDrivers_callback(GUID FAR* lpGUID, 
									 LPSTR lpDriverDesc, 
									 LPSTR lpDriverName, 
									 LPVOID lpContext )
{
//	LPD3DDEVICEDESC		lpD3DHWDevDesc; 
//	LPD3DDEVICEDESC		lpD3DHELDevDesc;
	CDevice				*pDevice;
	CVideo				*pVideo = (CVideo*) lpContext;
	
	pDevice = new CDevice( lpDriverName, lpDriverDesc );

	if (lpGUID != NULL)
	{
		memcpy( (void*)&pDevice->m_GUID, (void*)lpGUID, sizeof (GUID) );
		pDevice->m_DeviceNumber = DeviceNumber++;
	}
	else
	{
		pDevice->m_DeviceNumber = 1;		// Primary device!
		pDevice->m_pGUID = NULL;
	}

	// Link into list
	if (pVideo->m_pFirstDevice == NULL)
	{
		pVideo->m_pFirstDevice = pVideo->m_pLastDevice = pDevice;
	}
	else
	{
		pVideo->m_pLastDevice->m_pNext = pDevice;
		pVideo->m_pLastDevice = pDevice;
	}


	pVideo->GetDDrawHandle(pDevice);		// get handle for THIS device

//	memset( lpD3DHWDevDesc, 0, sizeof(DDCAPS) );
//	memset( lpD3DHELDevDesc,0, sizeof(DDCAPS) );
//	lpD3DHWDevDesc.dwSize	= sizeof(DDCAPS);
//	lpD3DHELDevDesc.dwSize	= sizeof(DDCAPS);
//	pVideo->m_DDraw.pDD4->GetCaps( lpD3DHWDevDesc,  lpD3DHELDevDesc );
//	pDevice->VideoMemory=pHardware->DDDriverCaps.dwVidMemTotal;
 

	pVideo->m_DDraw.pDD4->EnumDisplayModes(	0, NULL, pDevice, EnumModesCallback );

	pVideo->ReleaseDDrawHandle();			// release handle
	return 1;		// carry on;
}





// ******************************************************************
//
//	This version is used to enumerate under MULTI-monitor systems.
//	But, we just chain onto the older one anyway. :-)
//
// ******************************************************************
BOOL WINAPI DDEnumCallbackEx(GUID FAR *lpGUID, 
							 LPSTR     lpDriverDescription, 
							 LPSTR     lpDriverName,        
							 LPVOID    lpContext,           
							 HMONITOR  hm        
							)
{
	return EnumDrivers_callback(lpGUID, lpDriverDescription, lpDriverName, lpContext);
}
 

// ******************************************************************
//
//	Video constructor
//
// ******************************************************************
CVideo::CVideo( HINSTANCE hInstance )
{
	m_flags			= 0;
	m_hInstance		= hInstance;
	m_pFirstDevice	= NULL;
	m_pLastDevice	= NULL;
	m_err			= 0;
	m_pErr			= NULL;
	
	m_CurrentDevice	= 0;
	m_pCurrentDevice= NULL;
	m_CurrentMode	= 0;

	m_lPitch		= 0;
	m_pAddress		= NULL;

	m_DDraw.pDD4	 = NULL;
	m_DDraw.pDD		 = NULL;
	m_DDraw.pPrimary = NULL;
	m_DDraw.pBack	 = NULL;
	m_DDraw.pClipper = NULL;

	memset( &m_DDraw.ddsd, 0, sizeof(m_DDraw.ddsd) );
}





// ******************************************************************
//
//	Init Video system, enumerate all devices and screen modes.
//
// ******************************************************************
int32	CVideo::Init( uint32 flags )
{
    LPDIRECTDRAWENUMERATEEX lpDDEnumEx;
    HINSTANCE	h;
	
	NumScreenModes	= 1;					// Start screen mode number (0 is unused)
	DeviceNumber	= 2;					// Starts at 2 since 1 is reserved for Primary Device
	
	h = LoadLibrary("ddraw.dll");			// Load DDRAW.DLL, if we can't find it
    if (!h) return MD_ERROR;				// ddraw isn't installed.

	// Now look for the Multi-Monitor enumerate callback.
    lpDDEnumEx = (LPDIRECTDRAWENUMERATEEX) GetProcAddress(h,"DirectDrawEnumerateExA");
 

	// If its there, use THAT to enumerate, otherwise use the older one.
	if (lpDDEnumEx)
	{
        lpDDEnumEx( DDEnumCallbackEx, this, DDENUM_ATTACHEDSECONDARYDEVICES |DDENUM_NONDISPLAYDEVICES );
	}
	else		
	{
        DirectDrawEnumerate( EnumDrivers_callback, this );		// New call isn't found, so use old call.
	}


    FreeLibrary(h);			// Free DDraw dll.
	return MD_OKAY;
}


// ******************************************************************
//
//	Find a more based on width, height and bitdepth, and return a 
//	screen number
//
// ******************************************************************
int32	CVideo::CheckMode( uint32 width, uint32 height, uint32 bpp )
{
	CDevice		*pDevice;
	CScreenMode	*pScreenMode;

	pDevice = m_pFirstDevice;
	while (pDevice != NULL)
	{	
		// If NO device has been selected, search them ALL! Otherwise search 
		// ONLY the device the user has selected.
		if ((m_CurrentDevice == 0) OR (pDevice->m_DeviceNumber == m_CurrentDevice))
		{
			pScreenMode = pDevice->m_pFirstMode;
			while (pScreenMode != NULL)
			{
				if ((pScreenMode->m_xsize == width) AND 
					(pScreenMode->m_ysize == height) AND 
					(pScreenMode->m_bpp	  == bpp))
				{
					return pScreenMode->m_ModeNumber;
				}
				pScreenMode = pScreenMode->m_pNext;
			}
		}
		pDevice = pDevice->m_pNext;
	}
	return 0;
}


// ******************************************************************
//
//	Given a Mode Number, return the Screen class assosiated with it.
//
// ******************************************************************
CScreenMode	*CVideo::FindMode( int32 ModeNumber )
{
	CDevice		*pDevice;
	CScreenMode	*pScreenMode;

	pDevice = m_pFirstDevice;
	while (pDevice != NULL)
	{	
		pScreenMode = pDevice->m_pFirstMode;
		while (pScreenMode != NULL)
		{
			if( pScreenMode->m_ModeNumber == ModeNumber)
				return pScreenMode;

			pScreenMode = pScreenMode->m_pNext;
		}
		pDevice = pDevice->m_pNext;
	}
	return 0;
}

// ******************************************************************
//
// Given a device ID, find the class assosiated with it.
//
// ******************************************************************
CDevice *CVideo::FindDevice( int32 deviceid )
{
	CDevice	*pDevice;

	pDevice = m_pFirstDevice;
	while (pDevice != NULL)
	{
		if( pDevice->m_DeviceNumber == deviceid )
		{
			return pDevice;
		}
		pDevice = pDevice->m_pNext;
	}
	return NULL;
}


// ******************************************************************
//
//	Set the current device, if "0" just free the device.
//
// ******************************************************************
int CVideo::SetDevice( int32 device )
{
	CDevice	*pDevice = NULL;

	if( m_pFirstDevice == NULL ) return MD_ERROR;	// NO devices!! Probably NOT initialised, or ShutDown.

	// If the device that has been requested is the same as 
	// the one currently set, just return okay...
	if (m_CurrentDevice == device) return MD_OKAY;


	// If we have another device selected, free that first.
	if (m_CurrentDevice != 0)
	{
		CloseScreen();						// Make sure we don't have a screen open.
		ReleaseDDrawHandle();				// release handle if we HAVE one!
		m_CurrentDevice  = 0;				// Set NO device not active
		m_pCurrentDevice = NULL;			// the pointer to the ACTUAL device class
	}


	// if we just wanted to free the device, we can return now.
	if (device == 0) return MD_OKAY;
	pDevice = FindDevice( device );			// Now search for the correct device.
	if (pDevice == NULL) return MD_ERROR;	// can't find the device!!!


	// Now we have found the device we want, get a DirectDraw handle for it.
	// if theres a problem getting the device's DirectDraw handle, return an error.
	if (GetDDrawHandle( pDevice ) != MD_OKAY) return MD_ERROR;
	m_CurrentDevice  = device;				// set the current device, and
	m_pCurrentDevice = pDevice;				// the pointer to the ACTUAL device class

	return MD_OKAY;							// all okay! return.
}

// ******************************************************************
//
//  Check to see if a screen mode is open, and if so shut it down.
//
// ******************************************************************
int	CVideo::CloseScreen( void )
{
	if( m_CurrentMode == 0 ) return MD_ERROR;
	
	// Do we have a DDraw Handle?
	if( m_DDraw.pDD4 != NULL )
	{
		if( m_DDraw.pPrimary != NULL )
		{
			m_err = m_DDraw.pDD4->RestoreDisplayMode();		// get desktop back
			if( m_err != DD_OK)
			{
				m_pErr = GetError( m_err );		// find error message
				return MD_ERROR;				// return an error.
			}
			m_err = m_DDraw.pDD4->SetCooperativeLevel( m_MainWindow, DDSCL_NORMAL );
			if( m_err != DD_OK)
			{
				m_pErr = GetError( m_err );		// find error message
				return MD_ERROR;				// return an error.
			}
			
			m_err = m_DDraw.pPrimary->Release( );	// just returns the current counter..
			if( m_CurrentMode == VIDEO_WINDOWED )
			{
				m_err = m_DDraw.pBack->Release( );	// release seprate surface used in WINDOWED mode
			}

			m_DDraw.pPrimary = NULL;
			m_DDraw.pBack	 = NULL;
			m_CurrentMode	 = 0;
		}
	}

	return MD_OKAY;
}

// ******************************************************************
//
// Open a DirectDraw in WINDOWED mode.
//
// ******************************************************************
int CVideo::OpenWindowedMode( HWND Window, int32 Mode, int32 flags )
{
	RECT		ClientRect;
	CScreenMode	*pScreenMode;	

	
	// First, check to see if we have a device selected, if not select the PRIMARY device.
	if (m_CurrentDevice == 0) 
	{
		if( SetDevice( VIDEO_PRIMARYDEVICE ) != MD_OKAY ) return MD_OKAY;
	}


	// First we need to set the cooperative level, in WINDOWED mode, this is "normal"
	m_err = m_DDraw.pDD4->SetCooperativeLevel( m_MainWindow, DDSCL_NORMAL );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}


	// Next we will create a primary surface. First we need to
	// clear the DDSURFACEDESC2 structure and set the flags we need 
	// for the primary surface.
	memset( &m_DDraw.ddsd, 0, sizeof(m_DDraw.ddsd) );
	m_DDraw.ddsd.dwSize	 = sizeof(m_DDraw.ddsd);
	m_DDraw.ddsd.dwFlags = DDSD_CAPS;
	m_DDraw.ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	m_err = m_DDraw.pDD4->CreateSurface(  &m_DDraw.ddsd, &m_DDraw.pPrimary, NULL );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}



	// Now we must create a CLIPPER for this surface, then we need 
	// to tell the clipper which window to obtain the clipping info from,
	// and finally we must attach the clipper to the surface
	m_err = m_DDraw.pDD4->CreateClipper(  0, &m_DDraw.pClipper, NULL );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}
	// set clipper window.
	m_err = m_DDraw.pClipper->SetHWnd( 0, m_MainWindow );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}
	// attach clipper to the primary surface.
	m_err = m_DDraw.pPrimary->SetClipper( m_DDraw.pClipper );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}


	// Next we will getthe size of the window, and create a backbuffer for it.
	GetClientRect( m_MainWindow, &ClientRect);


	// Set up backbuffer data.
	memset( &m_DDraw.ddsd, 0, sizeof( m_DDraw.ddsd ));
	m_DDraw.ddsd.dwSize   = sizeof( m_DDraw.ddsd );
	m_DDraw.ddsd.dwFlags  = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;	// set valid feilds
	m_DDraw.ddsd.dwWidth  = ClientRect.right-ClientRect.left;
	m_DDraw.ddsd.dwHeight = ClientRect.bottom-ClientRect.top;
	
	// Now we want to process other options, like 3d and videoram enabled
	if( (flags & VIDEO_VIDEORAM) !=0 )
	{
		m_DDraw.ddsd.ddsCaps.dwCaps = DDSCAPS_VIDEOMEMORY;		// put surface in VIDEO memory
		m_flags |= VIDEO_SYS_VIDEORAM;					// surface is in video ram
	}
	else
	{
		m_DDraw.ddsd.ddsCaps.dwCaps = DDSCAPS_SYSTEMMEMORY;		// put surface in VIDEO memory
		m_flags &= ~VIDEO_SYS_VIDEORAM;							// surface is NOT in video ram

	}

	// If user wants a mode able to do 3D, force correct options...
	if( (flags & VIDEO_3D) != 0)
	{
		m_DDraw.ddsd.ddsCaps.dwCaps &= ~DDSCAPS_SYSTEMMEMORY;	// MUST be in VIDEO memory
		m_DDraw.ddsd.ddsCaps.dwCaps = DDSCAPS_3DDEVICE | DDSCAPS_VIDEOMEMORY;
		m_flags |= VIDEO_SYS_VIDEORAM;					// surface is in video ram
	}



	// Now, make up the back buffer! This will go in either System, or Video memory
	// depending on what options the user has chosen. If for some reason this fails,
	// we can't try anything else since if the user wanted it in VIDEORAM for some reason,
	// OR wanted a 3D capable surface then it can't exist in any other place.
	m_err = m_DDraw.pDD4->CreateSurface(  &m_DDraw.ddsd, &m_DDraw.pBack, NULL );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}


	// Now the mode has been created, we can get the pixel shift data we need
	memset( &m_DDraw.ddsd, 0, sizeof( m_DDraw.ddsd ));
	m_DDraw.ddsd.dwSize = sizeof( m_DDraw.ddsd );

	// Get surface pixel format details, This will in effect tell us the format 
	// of the Desktop.
	m_err = m_DDraw.pBack->GetSurfaceDesc( &m_DDraw.ddsd );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}

	pScreenMode		= new CScreenMode( &m_DDraw.ddsd );
	m_CurrentMode	= Mode;
	m_xsize			= pScreenMode->m_xsize;
	m_ysize			= pScreenMode->m_ysize;
	m_bpp			= pScreenMode->m_bpp;
	m_RedAND		= pScreenMode->m_RedAND;
	m_RedShift		= pScreenMode->m_RedShift;
	m_GreenAND		= pScreenMode->m_GreenAND;
	m_GreenShift	= pScreenMode->m_GreenShift;
	m_BlueAND		= pScreenMode->m_BlueAND;
	m_BlueShift		= pScreenMode->m_BlueShift;

	delete pScreenMode;
	return MD_OKAY;
}


// ******************************************************************
//
// Open a DirectDraw WINDOWED mode.
//
// ******************************************************************
int CVideo::OpenFullScreenMode( HWND Window, int32 Mode, int32 flags )
{
	DDSCAPS2	ddscaps;
	CScreenMode	*pScreenMode;	

	pScreenMode = FindMode( Mode );				// Find requested mode. This doesn't look at device numbers...
	if( pScreenMode == NULL ) return MD_ERROR;	// if we couldn't find it... return error.	
	if( SetDevice ( pScreenMode->m_Device ) != MD_OKAY) return MD_ERROR;	// set the device we need, or return error



	// Now set the exclusive mode...(and allow reboot in debug mode)
#ifdef _DEBUG
	m_err = m_DDraw.pDD4->SetCooperativeLevel( m_MainWindow, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWMODEX | DDSCL_ALLOWREBOOT );
#else
	m_err = m_DDraw.pDD4->SetCooperativeLevel( m_MainWindow, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWMODEX );
#endif
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}
	



	// I foundthe requested mode, so now we set it...
	m_err = m_DDraw.pDD4->SetDisplayMode(	pScreenMode->m_xsize,
											pScreenMode->m_ysize,
											pScreenMode->m_bpp,
											0,
											0);
	if( m_err != DD_OK)					// Can't set Mode? Bugger...
	{
		m_pErr = GetError( m_err );		
		return MD_ERROR;				
	}


	// Now, we set up caps for ddraw surface.
	m_flags &= ~VIDEO_SYS_DOUBLEBUFFER;
	m_flags |= (VIDEO_SYS_TRIPLEBUFFER | VIDEO_SYS_VIDEORAM);
	memset( &m_DDraw.ddsd, 0, sizeof(m_DDraw.ddsd) );
	m_DDraw.ddsd.dwSize				= sizeof(m_DDraw.ddsd);
	m_DDraw.ddsd.dwFlags			= DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	m_DDraw.ddsd.dwBackBufferCount	= 2;
	m_DDraw.ddsd.ddsCaps.dwCaps		=	DDSCAPS_PRIMARYSURFACE |
										DDSCAPS_FLIP |
										DDSCAPS_COMPLEX |
										DDSCAPS_VIDEOMEMORY;
	if( (flags&VIDEO_3D)!=0 )
	{
		m_DDraw.ddsd.ddsCaps.dwCaps	|=	DDSCAPS_3DDEVICE;
	}


	//
	// Now we try a cascade of buffer creation, from triple down to single!
	//
	m_err = ~DD_OK;						// by default we Don't want this mode
	if ( (flags&VIDEO_TRIPLEBUFFER)!=0 )
	{
		m_err = m_DDraw.pDD4->CreateSurface( &m_DDraw.ddsd, &m_DDraw.pPrimary, NULL );
	}


	// Did we fail to make a triple buffer? Try for a double buffer instead!
	if( m_err != DD_OK )
	{
		m_pErr = GetError( m_err );		// find error message

		m_flags &= ~VIDEO_SYS_TRIPLEBUFFER;
		m_flags |= VIDEO_SYS_DOUBLEBUFFER;
		m_DDraw.ddsd.dwBackBufferCount = 1;
		if( (flags&VIDEO_DOUBLEBUFFER) OR (flags&VIDEO_TRIPLEBUFFER) )
		{
			m_err = m_DDraw.pDD4->CreateSurface( &m_DDraw.ddsd, &m_DDraw.pPrimary, NULL );
		}



		// if we didn't manage to open this either, try a single buffer.
		if( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
	

			// Pooh, we never managed to create a double buffer in VRAM, so last
			// try, we'll go for a software double buffer.
			m_flags &= ~(VIDEO_SYS_DOUBLEBUFFER|VIDEO_SYS_VIDEORAM);		// surface is NOT in video ram anymore. 
			if( (flags&VIDEO_VIDEORAM)!=0) return MD_ERROR;					// if we NEEDED vram, return error.
			m_DDraw.ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;



			m_err = m_DDraw.pDD4->CreateSurface( &m_DDraw.ddsd, &m_DDraw.pPrimary, NULL );
			if( m_err != DD_OK )
			{
				m_pErr = GetError( m_err );		// find error message
				return MD_ERROR;
			}
		}

	}

	// If we get here, we've managed to open a screen mode. Now we need to get 
	// the backbuffer surface pointer, and the surface pixel details..
	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
	m_err = m_DDraw.pPrimary->GetAttachedSurface( &ddscaps, &m_DDraw.pBack );
	if( m_err != DD_OK )
	{
		m_pErr = GetError( m_err );		// No idea what could go wrong here!
		return MD_ERROR;
	}
	

	// get Surface description and then get pixel shift details..
	memset( &m_DDraw.ddsd, 0, sizeof( m_DDraw.ddsd ));
	m_DDraw.ddsd.dwSize = sizeof( m_DDraw.ddsd );
	m_err = m_DDraw.pBack->GetSurfaceDesc( &m_DDraw.ddsd );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}
	
	
	pScreenMode		= new CScreenMode( &m_DDraw.ddsd );
	m_CurrentMode	= Mode;
	m_xsize			= pScreenMode->m_xsize;
	m_ysize			= pScreenMode->m_ysize;
	m_bpp			= pScreenMode->m_bpp;
	m_RedAND		= pScreenMode->m_RedAND;
	m_RedShift		= pScreenMode->m_RedShift;
	m_GreenAND		= pScreenMode->m_GreenAND;
	m_GreenShift	= pScreenMode->m_GreenShift;
	m_BlueAND		= pScreenMode->m_BlueAND;
	m_BlueShift		= pScreenMode->m_BlueShift;
	
	delete pScreenMode;

	return MD_OKAY;
}


// ******************************************************************
//
// Function:	Setup a display mode, windowed of full screen.
//
// In:			Window	= Window handle
//				Mode	= display mode or VIDEO_WINDOWED for windowed mode
//				flags	= VIDEO_3D, VIDEO_VIDEORAM etc.
//
// Out:			MD_OKAY or MD_ERROR
//				on MD_ERROR m_err and m_pErr hold error codes.
//
// ******************************************************************
int	CVideo::SetMode( HWND Window, int32 Mode, int32 flags )
{
	// Show and update window
	UpdateWindow( Window );
	m_MainWindow = Window;

	// If we have a mode OPEN, close it before we open a NEW mode.
	if( m_CurrentMode != 0 )	CloseScreen();


	// Now open the correct video mode, either in a WINDOW on the desktop,
	// or in full screen sxclusive mode.
	if ( Mode == VIDEO_WINDOWED )
	{
		return OpenWindowedMode( Window, Mode, flags );
	}
	else	
	{
		return OpenFullScreenMode( Window, Mode, flags );
	}
}


// ******************************************************************
//
//	Flip. In windowed mode this will to a StreachBlit, and
//	in full screen exclusive mode, it will flip screens.
//
// ******************************************************************
int	CVideo::Flip( void )
{
	if (m_DDraw.pPrimary == NULL) return MD_ERROR;
	if (m_DDraw.pBack == NULL) return MD_ERROR;


	// First, we need to check and see if any of our surfaces have been lost, and if so
	// we have to restore them before we flip buffers. First check the Primary surface
	m_flags &= ~(VIDEO_SYS_BACKLOST | VIDEO_SYS_PRIMARYLOST);		// not lost..YET!
	m_err = m_DDraw.pPrimary->IsLost();
	if ( m_err == DDERR_SURFACELOST )
	{
		m_flags |= VIDEO_SYS_PRIMARYLOST;
		m_err = m_DDraw.pPrimary->Restore();					// try and get it back
		if ( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
		}
	}

	// Now check the Back Buffer
	m_err = m_DDraw.pBack->IsLost();
	if ( m_err == DDERR_SURFACELOST )
	{
		m_flags |= VIDEO_SYS_BACKLOST;
		m_err = m_DDraw.pBack->Restore();					// try and get it back
		if ( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
		}
	}

	
	
	// Check which mode, and blit or flip accordingly.
	if( m_CurrentMode == VIDEO_WINDOWED )
	{
		RECT	Dest,Source;
		POINT	Coord;

		Source.top	  = 0;
		Source.bottom = m_ysize;
		Source.left	  = 0;
		Source.right  = m_xsize;

		GetClientRect( m_MainWindow, &Dest );		// get blit TO size
		Coord.x = Dest.left;
		Coord.y = Dest.top;
		ClientToScreen( m_MainWindow, &Coord );		// get top left of window, minus menus and frame etc.
		Dest.left	+= Coord.x;						// move Dest window to new location
		Dest.right	+= Coord.x;
		Dest.top	+= Coord.y;
		Dest.bottom	+= Coord.y;

		m_err = m_DDraw.pPrimary->Blt( &Dest, m_DDraw.pBack, &Source, DDBLT_WAIT, NULL );
		if ( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
		}
		return MD_OKAY;						// flipped just fine!
	}
	else
	{
		m_DDraw.pPrimary->Flip( NULL, DDFLIP_WAIT );
	}
	return MD_ERROR;						// flipped just fine!
}

// ******************************************************************
//
//	ClearScreen: Clear the screen with a user supplied colour.
//
// ******************************************************************
int	CVideo::ClearScreen( uint32 ARGB, int x1, int y1, int x2, int y2)
{
	uint8	Mask[9]={0,1,3,7,15,31,63,127,255};
	RECT	rcDest;
	DDBLTFX bltFX;
	uint32	Red,Green,Blue;

	Red		= ARGB&m_RedAND;
	Green	= ARGB&m_GreenAND;
	Blue	= ARGB&m_BlueAND;
	Red		= Red>>m_RedShift;
	Green	= Green>>m_GreenShift;
	Blue	= Blue>>m_BlueShift;

	// Default is clear FULL screen
	if (x1>x2){
		x1 = 0;
		y1 = 0;
		x2 = m_xsize;
		y2 = m_ysize;
	}

	rcDest.left = x1;
	rcDest.top = y1;
	rcDest.right = x2;
	rcDest.bottom = y2;


	bltFX.dwSize = sizeof( bltFX );
	bltFX.dwFillColor = Red|Green|Blue;


	m_err = m_DDraw.pBack->IsLost();
	if ( m_err == DDERR_SURFACELOST )
	{
		m_flags |= VIDEO_SYS_BACKLOST;
		m_err = m_DDraw.pBack->Restore();	// try and get it back
		if ( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
		}
		rcDest.left = 0;					// if the surface is lost, clear the WHOLE screen!
		rcDest.top = 0;
		rcDest.right = m_xsize;
		rcDest.bottom = m_ysize;
	}

	m_err = m_DDraw.pBack->Blt( &rcDest, NULL, NULL, DDBLT_WAIT|DDBLT_COLORFILL, &bltFX );
	if (m_err != DD_OK)
	{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
	}
	return MD_OKAY;
}


// ******************************************************************
//
//	ShutDown: Close down any screen modes, and free DDraw Handles.
//	Also free any and all found screen modes and devices.
//
// ******************************************************************
void CVideo::ShutDown( void )
{
	CDevice		*pDevice,*pNextDevice;
	CScreenMode	*pScreenMode,*pNextScreenMode;

	SetDevice( 0 );							// free any selected device (also closes any open screen mode)

	// Free all devices and screen modes found.
	pDevice = m_pFirstDevice;
	while (pDevice != NULL)
	{	
		pNextDevice = pDevice->m_pNext;
		pScreenMode = pDevice->m_pFirstMode;
		while (pScreenMode != NULL)
		{
			pNextScreenMode = pScreenMode->m_pNext;
			delete pScreenMode;
			pScreenMode = pNextScreenMode ;
		}
		delete pDevice;
		pDevice = pNextDevice;
	}
	m_pFirstDevice = NULL;
	m_pLastDevice  = NULL;
}

// ******************************************************************
//
//	Video DEstructor. If a video mode is still open, the window
//  MUST not be Destroyed before calling this, of ShutDown() will
//  fail.
//
// ******************************************************************
CVideo::~CVideo( void )
{
	ShutDown();
}


// ******************************************************************
//
//	Lock Back buffer
//
// ******************************************************************
int32	CVideo::Lock( void )
{
	if( m_DDraw.pBack == NULL ) return MD_ERROR;			// we don't have a buffer to lock.
	if( (m_flags&VIDEO_SYS_LOCKED) != 0) return MD_ERROR;	// Surface is already locked.

	// Have we lost the back buffer?
	m_err = m_DDraw.pBack->IsLost();
	if ( m_err == DDERR_SURFACELOST )
	{
		m_flags |= VIDEO_SYS_PRIMARYLOST;
		m_err = m_DDraw.pBack->Restore();					// try and get it back
		if ( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
		}
	}

	memset( &m_DDraw.ddsd, 0, sizeof(m_DDraw.ddsd) );
	m_DDraw.ddsd.dwSize = sizeof(m_DDraw.ddsd);

	m_err = m_DDraw.pBack->Lock( NULL, &m_DDraw.ddsd, DDLOCK_WAIT, NULL);
	if ( m_err != DD_OK )
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}
	m_pAddress	= (uint8*) m_DDraw.ddsd.lpSurface;		// get address of back buffer.
	m_lPitch	= m_DDraw.ddsd.lPitch;			// get pitch of surface

	m_flags |= VIDEO_SYS_LOCKED;		// surface is now locked
	return MD_OKAY;
}



// ******************************************************************
//
//	Unlock Back buffer
//
// ******************************************************************
int32	CVideo::Unlock( void )
{
	if( m_DDraw.pBack == NULL ) return MD_ERROR;			// we don't have a buffer to lock.
	if( (m_flags&VIDEO_SYS_LOCKED) == 0) return MD_ERROR;	// Surface is already locked.

	m_err = m_DDraw.pBack->Unlock( NULL );
	if ( m_err != DD_OK )
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}

	m_pAddress	= NULL;
	m_lPitch	= 0;

	m_flags &= ~VIDEO_SYS_LOCKED;		// surface is now unlocked
	return MD_OKAY;
}

//***************************************************************
//																*
// Calculate 256 gamma look up table entries.					*
//																*
//	LUT value = Normalized LUT value ^ (1/Gamma)				*
//	By Dr. Gary Bishop,											*
//	http://wwwx.cs.unc.edu/~hoff/projects/comp235/gamma_dither/assign2.html		*
//																*
//***************************************************************
void CVideo::ComputeGammaCorrectedLUTentries(uint16 *pLUT, float Gamma)
{
	float NormInputV = 0.0f;
	float InputVincr = 1.0f / 255.0f;
	int	LUTvalue;

	Gamma = 1.0f/Gamma;
	for (LUTvalue=0; LUTvalue < 256; LUTvalue++)
	{
		float GammaCorrNormInputV = (float)pow(NormInputV, Gamma);
		int NewLUTvalue = (int)(GammaCorrNormInputV * (float)65535.0f);

		pLUT[LUTvalue]=(uint16)NewLUTvalue;
		NormInputV += InputVincr;
	}
}

// ******************************************************************
//
//	SetGamma(R,G,B): all are 0 to ???. 1.0f being "normal"
//
// ******************************************************************
int32 CVideo::SetGamma( float RedGamma, float GreenGamma, float BlueGamma )
{
	IDirectDrawGammaControl	*pGamma;
	DDGAMMARAMP				ddGammaRamp;

	m_err  = m_DDraw.pPrimary->QueryInterface( IID_IDirectDrawGammaControl, (void**)&pGamma );
	if ( m_err != DD_OK )
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}


	m_err = pGamma->GetGammaRamp( 0, &ddGammaRamp );
	if ( m_err != DD_OK )
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}

	ComputeGammaCorrectedLUTentries( &ddGammaRamp.red[0], RedGamma );
	ComputeGammaCorrectedLUTentries( &ddGammaRamp.green[0], GreenGamma );
	ComputeGammaCorrectedLUTentries( &ddGammaRamp.blue[0], BlueGamma );

	m_err = pGamma->SetGammaRamp( 0, &ddGammaRamp );
	if ( m_err != DD_OK )
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}

	return MD_OKAY;
}

#else	//__NT_DDRAW__
//******************************************************************************************************************
//******************************************************************************************************************
//******************************************************************************************************************
//******************************************************************************************************************


// ******************************************************************
//
//	Create a Device class
//
// ******************************************************************
CDevice::CDevice( char *pName, char *pDesc )
{
	memset( &m_GUID, 0, sizeof(GUID) );			// clear GUID (this means primary display)

	m_pGUID	= &m_GUID;
	m_pNext = NULL;
	m_pDriverName = (char*) MALLOC( Len(pName) + Len(pDesc) );
	m_pDriverDescription = m_pDriverName + Len(pName);
	m_pFirstMode = NULL;
	m_pLastMode	 = NULL;

	memcpy( m_pDriverName, pName, Len(pName) );
	memcpy( m_pDriverDescription, pDesc, Len(pDesc) );
}


// ******************************************************************
//
//	delete device class, and free
//
// ******************************************************************
CDevice::~CDevice( void )
{
	FREE( m_pDriverName );
}


// ******************************************************************
// Release the direct draw handles IF we have one.
// ******************************************************************
void CVideo::ReleaseDDrawHandle( void )
{
	// If we have an IDirectDraw Handle, free it!
	if (m_DDraw.pDD != NULL)
	{
		m_DDraw.pDD->Release();			// free handle
		m_DDraw.pDD = NULL;
	}
	m_CurrentDevice = 0;				// clear current device
}


// ******************************************************************
// Get direct draw handle of a passed device
// ******************************************************************
int CVideo::GetDDrawHandle( CDevice *pDevice )
{
	ReleaseDDrawHandle();			// release handle if we HAVE one!
	

	m_err = DirectDrawCreate(pDevice->m_pGUID, &m_DDraw.pDD, 0);
	if (m_err != DD_OK ) 
	{
		m_pErr = GetError( m_err );
		return MD_ERROR;
	}
	m_CurrentDevice = pDevice->m_DeviceNumber;
	return MD_OKAY;
}

// ******************************************************************
// Get the BITS in a mask
// ******************************************************************
uint32 GetNumberOfBits( DWORD dwMask )
{
    uint32 wBits = 0;
    while( dwMask )
    {
        dwMask = dwMask & ( dwMask - 1 );  
        wBits++;
    }
    return wBits;
}



// ******************************************************************
//
//	Create a screen mode class. Also takes a surface and extracts
//	Shift info, x and y sizes and BPP.
//
// ******************************************************************
CScreenMode::CScreenMode( LPDDSURFACEDESC lpDDSurfaceDesc )
{
	uint32	bits,downcount,bitcount,bitmask;

	m_xsize = lpDDSurfaceDesc->dwWidth;
	m_ysize = lpDDSurfaceDesc->dwHeight;
	m_bpp	= lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount;	// get bits in screen - RGB bits per pixel (4, 8, 16, 24, or 32). 
	m_pNext = NULL;

	m_AlphaAND   = 0;
	m_AlphaShift = 0;


	// Is this an RGB mode?
	if (((lpDDSurfaceDesc->ddpfPixelFormat.dwFlags&DDPF_RGB)==0) OR 
		(lpDDSurfaceDesc->ddpfPixelFormat.dwRBitMask == 0))
	{
		m_RedAND		= 0;
		m_RedShift		= 0;
		m_GreenAND		= 0;
		m_GreenShift	= 0;
		m_BlueAND		= 0;
		m_BlueShift		= 0;
		return;
	}

	// RED
	bits = lpDDSurfaceDesc->ddpfPixelFormat.dwRBitMask; 
	downcount = 0;
	while ((bits&1)==0)
	{
		bits = bits>>1;
		downcount++;
	}
	bitmask = bits;
	bitcount = GetNumberOfBits(bits);

	m_RedAND		= bitmask<<(16+(8-bitcount));
	m_RedShift		= (16+(8-bitcount)) - downcount;


	// GREEN
	bits = lpDDSurfaceDesc->ddpfPixelFormat.dwGBitMask; 
	downcount = 0;
	while ((bits&1)==0)
	{
		bits = bits>>1;
		downcount++;
	}
	bitmask = bits;
	bitcount = GetNumberOfBits(bits);

	m_GreenAND		= bitmask<<(8+(8-bitcount));
	m_GreenShift	= (8+(8-bitcount)) - downcount;


	// BLUE
	bits = lpDDSurfaceDesc->ddpfPixelFormat.dwBBitMask; 
	downcount = 0;
	while ((bits&1)==0)
	{
		bits = bits>>1;
		downcount++;
	}
	bitmask = bits;
	bitcount = GetNumberOfBits(bits);

	m_BlueAND		= bitmask<<(8-bitcount);
	m_BlueShift		= (8-bitcount) - downcount;


	m_AlphaAND	 = 0;
	m_AlphaShift = 0;
}


// ******************************************************************
//
//	Screen Mode(s) Enumeration.
//
// ******************************************************************
HRESULT WINAPI EnumModesCallback( LPDDSURFACEDESC lpDDSurfaceDesc,    LPVOID lpContext )
{
	CDevice		*pDevice = (CDevice *) lpContext;
	CScreenMode	*pScreen;


	pScreen = new CScreenMode( lpDDSurfaceDesc );
	pScreen->m_Device = pDevice->m_DeviceNumber;		// set device number
	pScreen->m_ModeNumber = NumScreenModes;				// give made up screen number
	NumScreenModes++;

	if (pDevice->m_pFirstMode==NULL)
	{
		pDevice->m_pFirstMode = pDevice->m_pLastMode = pScreen;
	}
	else 
	{
		pDevice->m_pLastMode->m_pNext = pScreen;
		pDevice->m_pLastMode = pScreen;
	}

	return DDENUMRET_OK;		// carry on...
}


// ******************************************************************
//
//	Display Driver enumeration.
//
// ******************************************************************
BOOL FAR PASCAL EnumDrivers_callback(GUID FAR* lpGUID, 
									 LPSTR lpDriverDesc, 
									 LPSTR lpDriverName, 
									 LPVOID lpContext )
{
	CDevice				*pDevice;
	CVideo				*pVideo = (CVideo*) lpContext;
	
	pDevice = new CDevice( lpDriverName, lpDriverDesc );

	if (lpGUID != NULL)
	{
		memcpy( (void*)&pDevice->m_GUID, (void*)lpGUID, sizeof (GUID) );
		pDevice->m_DeviceNumber = DeviceNumber++;
	}
	else
	{
		pDevice->m_DeviceNumber = 1;		// Primary device!
		pDevice->m_pGUID = NULL;
	}

	// Link into list
	if (pVideo->m_pFirstDevice == NULL)
	{
		pVideo->m_pFirstDevice = pVideo->m_pLastDevice = pDevice;
	}
	else
	{
		pVideo->m_pLastDevice->m_pNext = pDevice;
		pVideo->m_pLastDevice = pDevice;
	}


	pVideo->GetDDrawHandle(pDevice);		// get handle for THIS device
	pVideo->m_DDraw.pDD->EnumDisplayModes(	0, NULL, pDevice, EnumModesCallback );
	pVideo->ReleaseDDrawHandle();			// release handle
	return 1;		// carry on;
}




// ******************************************************************
//
//	Video constructor
//
// ******************************************************************
CVideo::CVideo( HINSTANCE hInstance )
{
	m_flags			= 0;
	m_hInstance		= hInstance;
	m_pFirstDevice	= NULL;
	m_pLastDevice	= NULL;
	m_err			= 0;
	m_pErr			= NULL;
	
	m_CurrentDevice	= 0;
	m_pCurrentDevice= NULL;
	m_CurrentMode	= 0;

	m_lPitch		= 0;
	m_pAddress		= NULL;

	m_DDraw.pDD		 = NULL;
	m_DDraw.pPrimary = NULL;
	m_DDraw.pBack	 = NULL;
	m_DDraw.pClipper = NULL;

	memset( &m_DDraw.ddsd, 0, sizeof(m_DDraw.ddsd) );
}





// ******************************************************************
//
//	Init Video system, enumerate all devices and screen modes.
//
// ******************************************************************
int32	CVideo::Init( uint32 flags )
{
    HINSTANCE	h;
	
	NumScreenModes	= 1;					// Start screen mode number (0 is unused)
	DeviceNumber	= 2;					// Starts at 2 since 1 is reserved for Primary Device
	
	h = LoadLibrary("ddraw.dll");			// Load DDRAW.DLL, if we can't find it
    if (!h) return MD_ERROR;				// ddraw isn't installed.

	DirectDrawEnumerate( EnumDrivers_callback, this );		// New call isn't found, so use old call.

    FreeLibrary(h);			// Free DDraw dll.
	return MD_OKAY;
}


// ******************************************************************
//
//	Find a more based on width, height and bitdepth, and return a 
//	screen number
//
// ******************************************************************
int32	CVideo::CheckMode( uint32 width, uint32 height, uint32 bpp )
{
	CDevice		*pDevice;
	CScreenMode	*pScreenMode;

	pDevice = m_pFirstDevice;
	while (pDevice != NULL)
	{	
		// If NO device has been selected, search them ALL! Otherwise search 
		// ONLY the device the user has selected.
		if ((m_CurrentDevice == 0) OR (pDevice->m_DeviceNumber == m_CurrentDevice))
		{
			pScreenMode = pDevice->m_pFirstMode;
			while (pScreenMode != NULL)
			{
				if ((pScreenMode->m_xsize == width) AND 
					(pScreenMode->m_ysize == height) AND 
					(pScreenMode->m_bpp	  == bpp))
				{
					return pScreenMode->m_ModeNumber;
				}
				pScreenMode = pScreenMode->m_pNext;
			}
		}
		pDevice = pDevice->m_pNext;
	}
	return 0;
}


// ******************************************************************
//
//	Given a Mode Number, return the Screen class assosiated with it.
//
// ******************************************************************
CScreenMode	*CVideo::FindMode( int32 ModeNumber )
{
	CDevice		*pDevice;
	CScreenMode	*pScreenMode;

	pDevice = m_pFirstDevice;
	while (pDevice != NULL)
	{	
		pScreenMode = pDevice->m_pFirstMode;
		while (pScreenMode != NULL)
		{
			if( pScreenMode->m_ModeNumber == ModeNumber)
				return pScreenMode;

			pScreenMode = pScreenMode->m_pNext;
		}
		pDevice = pDevice->m_pNext;
	}
	return 0;
}

// ******************************************************************
//
// Given a device ID, find the class assosiated with it.
//
// ******************************************************************
CDevice *CVideo::FindDevice( int32 deviceid )
{
	CDevice	*pDevice;

	pDevice = m_pFirstDevice;
	while (pDevice != NULL)
	{
		if( pDevice->m_DeviceNumber == deviceid )
		{
			return pDevice;
		}
		pDevice = pDevice->m_pNext;
	}
	return NULL;
}


// ******************************************************************
//
//	Set the current device, if "0" just free the device.
//
// ******************************************************************
int CVideo::SetDevice( int32 device )
{
	CDevice	*pDevice = NULL;

	if( m_pFirstDevice == NULL ) return MD_ERROR;	// NO devices!! Probably NOT initialised, or ShutDown.

	// If the device that has been requested is the same as 
	// the one currently set, just return okay...
	if (m_CurrentDevice == device) return MD_OKAY;


	// If we have another device selected, free that first.
	if (m_CurrentDevice != 0)
	{
		CloseScreen();						// Make sure we don't have a screen open.
		ReleaseDDrawHandle();				// release handle if we HAVE one!
		m_CurrentDevice  = 0;				// Set NO device not active
		m_pCurrentDevice = NULL;			// the pointer to the ACTUAL device class
	}


	// if we just wanted to free the device, we can return now.
	if (device == 0) return MD_OKAY;
	pDevice = FindDevice( device );			// Now search for the correct device.
	if (pDevice == NULL) return MD_ERROR;	// can't find the device!!!


	// Now we have found the device we want, get a DirectDraw handle for it.
	// if theres a problem getting the device's DirectDraw handle, return an error.
	if (GetDDrawHandle( pDevice ) != MD_OKAY) return MD_ERROR;
	m_CurrentDevice  = device;				// set the current device, and
	m_pCurrentDevice = pDevice;				// the pointer to the ACTUAL device class

	return MD_OKAY;							// all okay! return.
}

// ******************************************************************
//
//  Check to see if a screen mode is open, and if so shut it down.
//
// ******************************************************************
int	CVideo::CloseScreen( void )
{
	if( m_CurrentMode == 0 ) return MD_ERROR;
	
	// Do we have a DDraw Handle?
	if( m_DDraw.pDD  != NULL )
	{
		if( m_DDraw.pPrimary != NULL )
		{
			//m_err = m_DDraw.pDD4->RestoreDisplayMode();		// get desktop back
			m_err = m_DDraw.pDD->RestoreDisplayMode();		// get desktop back
			if( m_err != DD_OK)
			{
				m_pErr = GetError( m_err );		// find error message
				return MD_ERROR;				// return an error.
			}
			//m_err = m_DDraw.pDD4->SetCooperativeLevel( m_MainWindow, DDSCL_NORMAL );
			m_err = m_DDraw.pDD->SetCooperativeLevel( m_MainWindow, DDSCL_NORMAL );
			if( m_err != DD_OK)
			{
				m_pErr = GetError( m_err );		// find error message
				return MD_ERROR;				// return an error.
			}
			
			m_err = m_DDraw.pPrimary->Release( );	// just returns the current counter..
			if( m_CurrentMode == VIDEO_WINDOWED )
			{
				m_err = m_DDraw.pBack->Release( );	// release seprate surface used in WINDOWED mode
			}

			m_DDraw.pPrimary = NULL;
			m_DDraw.pBack	 = NULL;
			m_CurrentMode	 = 0;
		}
	}

	return MD_OKAY;
}

// ******************************************************************
//
// Open a DirectDraw in WINDOWED mode.
//
// ******************************************************************
int CVideo::OpenWindowedMode( HWND Window, int32 Mode, int32 flags )
{
	RECT		ClientRect;
	CScreenMode	*pScreenMode;	

	
	// First, check to see if we have a device selected, if not select the PRIMARY device.
	if (m_CurrentDevice == 0) 
	{
		if( SetDevice( VIDEO_PRIMARYDEVICE ) != MD_OKAY ) return MD_OKAY;
	}


	// First we need to set the cooperative level, in WINDOWED mode, this is "normal"
	m_err = m_DDraw.pDD->SetCooperativeLevel( m_MainWindow, DDSCL_NORMAL );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}


	// Next we will create a primary surface. First we need to
	// clear the DDSURFACEDESC2 structure and set the flags we need 
	// for the primary surface.
	memset( &m_DDraw.ddsd, 0, sizeof(m_DDraw.ddsd) );
	m_DDraw.ddsd.dwSize	 = sizeof(m_DDraw.ddsd);
	m_DDraw.ddsd.dwFlags = DDSD_CAPS;
	m_DDraw.ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	m_err = m_DDraw.pDD->CreateSurface(  &m_DDraw.ddsd, &m_DDraw.pPrimary, NULL );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}



	// Now we must create a CLIPPER for this surface, then we need 
	// to tell the clipper which window to obtain the clipping info from,
	// and finally we must attach the clipper to the surface
	m_err = m_DDraw.pDD->CreateClipper(  0, &m_DDraw.pClipper, NULL );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}
	// set clipper window.
	m_err = m_DDraw.pClipper->SetHWnd( 0, m_MainWindow );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}
	// attach clipper to the primary surface.
	m_err = m_DDraw.pPrimary->SetClipper( m_DDraw.pClipper );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}


	// Next we will getthe size of the window, and create a backbuffer for it.
	GetClientRect( m_MainWindow, &ClientRect);


	// Set up backbuffer data.
	memset( &m_DDraw.ddsd, 0, sizeof( m_DDraw.ddsd ));
	m_DDraw.ddsd.dwSize   = sizeof( m_DDraw.ddsd );
	m_DDraw.ddsd.dwFlags  = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;	// set valid feilds
	m_DDraw.ddsd.dwWidth  = ClientRect.right-ClientRect.left;
	m_DDraw.ddsd.dwHeight = ClientRect.bottom-ClientRect.top;
	
	// Now we want to process other options, like 3d and videoram enabled
	if( (flags & VIDEO_VIDEORAM) !=0 )
	{
		m_DDraw.ddsd.ddsCaps.dwCaps = DDSCAPS_VIDEOMEMORY;		// put surface in VIDEO memory
		m_flags |= VIDEO_SYS_VIDEORAM;					// surface is in video ram
	}
	else
	{
		m_DDraw.ddsd.ddsCaps.dwCaps = DDSCAPS_SYSTEMMEMORY;		// put surface in VIDEO memory
		m_flags &= ~VIDEO_SYS_VIDEORAM;							// surface is NOT in video ram

	}

	// If user wants a mode able to do 3D, force correct options...
	if( (flags & VIDEO_3D) != 0)
	{
		m_DDraw.ddsd.ddsCaps.dwCaps &= ~DDSCAPS_SYSTEMMEMORY;	// MUST be in VIDEO memory
		m_DDraw.ddsd.ddsCaps.dwCaps = DDSCAPS_3DDEVICE | DDSCAPS_VIDEOMEMORY;
		m_flags |= VIDEO_SYS_VIDEORAM;					// surface is in video ram
	}



	// Now, make up the back buffer! This will go in either System, or Video memory
	// depending on what options the user has chosen. If for some reason this fails,
	// we can't try anything else since if the user wanted it in VIDEORAM for some reason,
	// OR wanted a 3D capable surface then it can't exist in any other place.
	m_err = m_DDraw.pDD->CreateSurface(  &m_DDraw.ddsd, &m_DDraw.pBack, NULL );
	if( m_err != DD_OK)
	{
		// If we can't make a VRAM buffer, use system memory... Its slower but works better on NT
		m_flags &= ~VIDEO_SYS_VIDEORAM;					// surface is in video ram
		m_DDraw.ddsd.ddsCaps.dwCaps = DDSCAPS_SYSTEMMEMORY;		// put surface in VIDEO memory
		m_err = m_DDraw.pDD->CreateSurface(  &m_DDraw.ddsd, &m_DDraw.pBack, NULL );
		if( m_err != DD_OK)
		{
			m_pErr = GetError( m_err );		// find error message
			MessageBox(NULL,m_pErr, "DDraw Error",0);
			return MD_ERROR;				// return an error.
		}
	}


	// Now the mode has been created, we can get the pixel shift data we need
	memset( &m_DDraw.ddsd, 0, sizeof( m_DDraw.ddsd ));
	m_DDraw.ddsd.dwSize = sizeof( m_DDraw.ddsd );

	// Get surface pixel format details, This will in effect tell us the format 
	// of the Desktop.
	m_err = m_DDraw.pBack->GetSurfaceDesc( &m_DDraw.ddsd );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}

	pScreenMode		= new CScreenMode( &m_DDraw.ddsd );
	m_CurrentMode	= Mode;
	m_xsize			= pScreenMode->m_xsize;
	m_ysize			= pScreenMode->m_ysize;
	m_bpp			= pScreenMode->m_bpp;
	m_RedAND		= pScreenMode->m_RedAND;
	m_RedShift		= pScreenMode->m_RedShift;
	m_GreenAND		= pScreenMode->m_GreenAND;
	m_GreenShift	= pScreenMode->m_GreenShift;
	m_BlueAND		= pScreenMode->m_BlueAND;
	m_BlueShift		= pScreenMode->m_BlueShift;

	delete pScreenMode;
	return MD_OKAY;
}


// ******************************************************************
//
// Open a DirectDraw WINDOWED mode.
//
// ******************************************************************
int CVideo::OpenFullScreenMode( HWND Window, int32 Mode, int32 flags )
{
//	DDSCAPS2	ddscaps;
	DDSCAPS		ddscaps;
	CScreenMode	*pScreenMode;	

	pScreenMode = FindMode( Mode );				// Find requested mode. This doesn't look at device numbers...
	if( pScreenMode == NULL ) return MD_ERROR;	// if we couldn't find it... return error.	
	if( SetDevice ( pScreenMode->m_Device ) != MD_OKAY) return MD_ERROR;	// set the device we need, or return error



	// Now set the exclusive mode...(and allow reboot in debug mode)
#ifdef _DEBUG
	m_err = m_DDraw.pDD->SetCooperativeLevel( m_MainWindow, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWMODEX | DDSCL_ALLOWREBOOT );
#else
	m_err = m_DDraw.pDD->SetCooperativeLevel( m_MainWindow, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWMODEX );
#endif
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}
	



	// I foundthe requested mode, so now we set it...
	m_err = m_DDraw.pDD->SetDisplayMode(	pScreenMode->m_xsize,
											pScreenMode->m_ysize,
											pScreenMode->m_bpp);
	if( m_err != DD_OK)					// Can't set Mode? Bugger...
	{
		m_pErr = GetError( m_err );		
		return MD_ERROR;				
	}


	// Now, we set up caps for ddraw surface.
	m_flags &= ~VIDEO_SYS_DOUBLEBUFFER;
	m_flags |= (VIDEO_SYS_TRIPLEBUFFER | VIDEO_SYS_VIDEORAM);
	memset( &m_DDraw.ddsd, 0, sizeof(m_DDraw.ddsd) );
	m_DDraw.ddsd.dwSize				= sizeof(m_DDraw.ddsd);
	m_DDraw.ddsd.dwFlags			= DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	m_DDraw.ddsd.dwBackBufferCount	= 2;
	m_DDraw.ddsd.ddsCaps.dwCaps		=	DDSCAPS_PRIMARYSURFACE |
										DDSCAPS_FLIP |
										DDSCAPS_COMPLEX;// |
										//DDSCAPS_VIDEOMEMORY;
	if( (flags&VIDEO_3D)!=0 )
	{
		m_DDraw.ddsd.ddsCaps.dwCaps	|=	DDSCAPS_3DDEVICE;
	}


	//
	// Now we try a cascade of buffer creation, from triple down to single!
	//
	m_err = ~DD_OK;						// by default we Don't want this mode
	if ( (flags&VIDEO_TRIPLEBUFFER)!=0 )
	{
		m_err = m_DDraw.pDD->CreateSurface( &m_DDraw.ddsd, &m_DDraw.pPrimary, NULL );
	}


	// Did we fail to make a triple buffer? Try for a double buffer instead!
	if( m_err != DD_OK )
	{
		m_pErr = GetError( m_err );		// find error message

		m_flags &= ~VIDEO_SYS_TRIPLEBUFFER;
		m_flags |= VIDEO_SYS_DOUBLEBUFFER;
		m_DDraw.ddsd.dwBackBufferCount = 1;
		if( (flags&VIDEO_DOUBLEBUFFER) || (flags&VIDEO_TRIPLEBUFFER) )
		{
			m_err = m_DDraw.pDD->CreateSurface( &m_DDraw.ddsd, &m_DDraw.pPrimary, NULL );
		}



		// if we didn't manage to open this either, try a single buffer.
		if( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
	

			// Pooh, we never managed to create a double buffer in VRAM, so last
			// try, we'll go for a software double buffer.
			m_flags &= ~(VIDEO_SYS_DOUBLEBUFFER|VIDEO_SYS_VIDEORAM);		// surface is NOT in video ram anymore. 
			if( (flags&VIDEO_VIDEORAM)!=0) return MD_ERROR;					// if we NEEDED vram, return error.
			m_DDraw.ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;



			m_err = m_DDraw.pDD->CreateSurface( &m_DDraw.ddsd, &m_DDraw.pPrimary, NULL );
			if( m_err != DD_OK )
			{
				m_pErr = GetError( m_err );		// find error message
				return MD_ERROR;
			}
		}

	}

	// If we get here, we've managed to open a screen mode. Now we need to get 
	// the backbuffer surface pointer, and the surface pixel details..
	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
	m_err = m_DDraw.pPrimary->GetAttachedSurface( &ddscaps, &m_DDraw.pBack );
	if( m_err != DD_OK )
	{
		m_pErr = GetError( m_err );		// No idea what could go wrong here!
		return MD_ERROR;
	}
	

	// get Surface description and then get pixel shift details..
	memset( &m_DDraw.ddsd, 0, sizeof( m_DDraw.ddsd ));
	m_DDraw.ddsd.dwSize = sizeof( m_DDraw.ddsd );
	m_err = m_DDraw.pBack->GetSurfaceDesc( &m_DDraw.ddsd );
	if( m_err != DD_OK)
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}
	
	
	pScreenMode		= new CScreenMode( &m_DDraw.ddsd );
	m_CurrentMode	= Mode;
	m_xsize			= pScreenMode->m_xsize;
	m_ysize			= pScreenMode->m_ysize;
	m_bpp			= pScreenMode->m_bpp;
	m_RedAND		= pScreenMode->m_RedAND;
	m_RedShift		= pScreenMode->m_RedShift;
	m_GreenAND		= pScreenMode->m_GreenAND;
	m_GreenShift	= pScreenMode->m_GreenShift;
	m_BlueAND		= pScreenMode->m_BlueAND;
	m_BlueShift		= pScreenMode->m_BlueShift;
	
	delete pScreenMode;

	return MD_OKAY;
}


// ******************************************************************
//
// Function:	Setup a display mode, windowed of full screen.
//
// In:			Window	= Window handle
//				Mode	= display mode or VIDEO_WINDOWED for windowed mode
//				flags	= VIDEO_3D, VIDEO_VIDEORAM etc.
//
// Out:			MD_OKAY or MD_ERROR
//				on MD_ERROR m_err and m_pErr hold error codes.
//
// ******************************************************************
int	CVideo::SetMode( HWND Window, int32 Mode, int32 flags )
{
	// Show and update window
	UpdateWindow( Window );
	m_MainWindow = Window;

	// If we have a mode OPEN, close it before we open a NEW mode.
	if( m_CurrentMode != 0 )	CloseScreen();


	// Now open the correct video mode, either in a WINDOW on the desktop,
	// or in full screen sxclusive mode.
	if ( Mode == VIDEO_WINDOWED )
	{
		return OpenWindowedMode( Window, Mode, flags );
	}
	else	
	{
		return OpenFullScreenMode( Window, Mode, flags );
	}
}


// ******************************************************************
//
//	Flip. In windowed mode this will to a StreachBlit, and
//	in full screen exclusive mode, it will flip screens.
//
// ******************************************************************
int	CVideo::Flip( void )
{
	if (m_DDraw.pPrimary == NULL) return MD_ERROR;
	if (m_DDraw.pBack == NULL) return MD_ERROR;


	// First, we need to check and see if any of our surfaces have been lost, and if so
	// we have to restore them before we flip buffers. First check the Primary surface
	m_flags &= ~(VIDEO_SYS_BACKLOST | VIDEO_SYS_PRIMARYLOST);		// not lost..YET!
	m_err = m_DDraw.pPrimary->IsLost();
	if ( m_err == DDERR_SURFACELOST )
	{
		m_flags |= VIDEO_SYS_PRIMARYLOST;
		m_err = m_DDraw.pPrimary->Restore();					// try and get it back
		if ( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
		}
	}

	// Now check the Back Buffer
	m_err = m_DDraw.pBack->IsLost();
	if ( m_err == DDERR_SURFACELOST )
	{
		m_flags |= VIDEO_SYS_BACKLOST;
		m_err = m_DDraw.pBack->Restore();					// try and get it back
		if ( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
		}
	}

	
	
	// Check which mode, and blit or flip accordingly.
	if( m_CurrentMode == VIDEO_WINDOWED )
	{
		RECT	Dest,Source;
		POINT	Coord;

		Source.top	  = 0;
		Source.bottom = m_ysize;
		Source.left	  = 0;
		Source.right  = m_xsize;

		GetClientRect( m_MainWindow, &Dest );		// get blit TO size
		Coord.x = Dest.left;
		Coord.y = Dest.top;
		ClientToScreen( m_MainWindow, &Coord );		// get top left of window, minus menus and frame etc.
		Dest.left	+= Coord.x;						// move Dest window to new location
		Dest.right	+= Coord.x;
		Dest.top	+= Coord.y;
		Dest.bottom	+= Coord.y;

		m_err = m_DDraw.pPrimary->Blt( &Dest, m_DDraw.pBack, &Source, DDBLT_WAIT, NULL );
		if ( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
		}
		return MD_OKAY;						// flipped just fine!
	}
	else
	{
		m_DDraw.pPrimary->Flip( NULL, DDFLIP_WAIT );
	}
	return MD_ERROR;						// flipped just fine!
}

// ******************************************************************
//
//	ClearScreen: Clear the screen with a user supplied colour.
//
// ******************************************************************
int	CVideo::ClearScreen( uint32 ARGB, int x1, int y1, int x2, int y2)
{
	uint8	Mask[9]={0,1,3,7,15,31,63,127,255};
	RECT	rcDest;
	DDBLTFX bltFX;
	uint32	Red,Green,Blue;

	Red		= ARGB&m_RedAND;
	Green	= ARGB&m_GreenAND;
	Blue	= ARGB&m_BlueAND;
	Red		= Red>>m_RedShift;
	Green	= Green>>m_GreenShift;
	Blue	= Blue>>m_BlueShift;

	// Default is clear FULL screen
	if (x1>x2){
		x1 = 0;
		y1 = 0;
		x2 = m_xsize;
		y2 = m_ysize;
	}

	rcDest.left = x1;
	rcDest.top = y1;
	rcDest.right = x2;
	rcDest.bottom = y2;


	bltFX.dwSize = sizeof( bltFX );
	bltFX.dwFillColor = Red|Green|Blue;


	m_err = m_DDraw.pBack->IsLost();
	if ( m_err == DDERR_SURFACELOST )
	{
		m_flags |= VIDEO_SYS_BACKLOST;
		m_err = m_DDraw.pBack->Restore();	// try and get it back
		if ( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
		}
		rcDest.left = 0;					// if the surface is lost, clear the WHOLE screen!
		rcDest.top = 0;
		rcDest.right = m_xsize;
		rcDest.bottom = m_ysize;
	}

	m_err = m_DDraw.pBack->Blt( &rcDest, NULL, NULL, DDBLT_WAIT|DDBLT_COLORFILL, &bltFX );
	if (m_err != DD_OK)
	{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
	}
	return MD_OKAY;
}


// ******************************************************************
//
//	ShutDown: Close down any screen modes, and free DDraw Handles.
//	Also free any and all found screen modes and devices.
//
// ******************************************************************
void CVideo::ShutDown( void )
{
	CDevice		*pDevice,*pNextDevice;
	CScreenMode	*pScreenMode,*pNextScreenMode;

	SetDevice( 0 );							// free any selected device (also closes any open screen mode)

	// Free all devices and screen modes found.
	pDevice = m_pFirstDevice;
	while (pDevice != NULL)
	{	
		pNextDevice = pDevice->m_pNext;
		pScreenMode = pDevice->m_pFirstMode;
		while (pScreenMode != NULL)
		{
			pNextScreenMode = pScreenMode->m_pNext;
			delete pScreenMode;
			pScreenMode = pNextScreenMode ;
		}
		delete pDevice;
		pDevice = pNextDevice;
	}
	m_pFirstDevice = NULL;
	m_pLastDevice  = NULL;
}

// ******************************************************************
//
//	Video DEstructor. If a video mode is still open, the window
//  MUST not be Destroyed before calling this, of ShutDown() will
//  fail.
//
// ******************************************************************
CVideo::~CVideo( void )
{
	ShutDown();
}


// ******************************************************************
//
//	Lock Back buffer
//
// ******************************************************************
int32	CVideo::Lock( void )
{
	if( m_DDraw.pBack == NULL ) return MD_ERROR;			// we don't have a buffer to lock.
	if( (m_flags&VIDEO_SYS_LOCKED) != 0) return MD_ERROR;	// Surface is already locked.

	// Have we lost the back buffer?
	m_err = m_DDraw.pBack->IsLost();
	if ( m_err == DDERR_SURFACELOST )
	{
		m_flags |= VIDEO_SYS_PRIMARYLOST;
		m_err = m_DDraw.pBack->Restore();					// try and get it back
		if ( m_err != DD_OK )
		{
			m_pErr = GetError( m_err );		// find error message
			return MD_ERROR;				// return an error.
		}
	}

	memset( &m_DDraw.ddsd, 0, sizeof(m_DDraw.ddsd) );
	m_DDraw.ddsd.dwSize = sizeof(m_DDraw.ddsd);

	m_err = m_DDraw.pBack->Lock( NULL, &m_DDraw.ddsd, DDLOCK_WAIT, NULL);
	if ( m_err != DD_OK )
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}
	m_pAddress	= (uint8*) m_DDraw.ddsd.lpSurface;		// get address of back buffer.
	m_lPitch	= m_DDraw.ddsd.lPitch;			// get pitch of surface

	m_flags |= VIDEO_SYS_LOCKED;		// surface is now locked
	return MD_OKAY;
}



// ******************************************************************
//
//	Unlock Back buffer
//
// ******************************************************************
int32	CVideo::Unlock( void )
{
	if( m_DDraw.pBack == NULL ) return MD_ERROR;			// we don't have a buffer to lock.
	if( (m_flags&VIDEO_SYS_LOCKED) == 0) return MD_ERROR;	// Surface is already locked.

	m_err = m_DDraw.pBack->Unlock( NULL );
	if ( m_err != DD_OK )
	{
		m_pErr = GetError( m_err );		// find error message
		return MD_ERROR;				// return an error.
	}

	m_pAddress	= NULL;
	m_lPitch	= 0;

	m_flags &= ~VIDEO_SYS_LOCKED;		// surface is now unlocked
	return MD_OKAY;
}

//******************************************************************************************************************
//******************************************************************************************************************
//******************************************************************************************************************
//******************************************************************************************************************
#endif